#brailleDisplayDrivers/papenmeier_serial.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Author: Tobias Platen (nvda@lists.thm.de)

import time
import wx
import braille
from logHandler import log


import inputCore
import struct
import speech #for debugging

import serial
import config


import shlobj
import os
import config


#Control Flow
STX = 0x02 #Start of Text
ETX = 0x03 #End of Text

KEY_CHECK_INTERVAL = 10



def brl_auto_id(): 
	"""send auto id command to braille display"""
	return chr(STX)+'S'+chr(0)+chr(0)+chr(0)+chr(0)+chr(ETX)
	#send a bad packet to the braille display
	

def brl_out(offset, data):
	"""send data to braille display"""
	ret = struct.pack('BB', STX, 0x53) #STX,'S'        
	d2 = len(data)+7
	ret += struct.pack('BB', offset / 256, offset % 256)
	ret += struct.pack('BB', 0, d2 % 256)
	for d in data:
		ret += struct.pack('B', d)
	ret += struct.pack('B', ETX)
	return ret



def brl_poll(dev):
	"""read data from braille display"""
	ret =''
	while(dev.inWaiting()>0):
		ret +=dev.read(1)
	return ret

def brl_poll2(dev):
	"""read data from braille display, used by keypress handler"""
	ret =''
	blockcount=0
	if(dev.inWaiting()==0): return ret
	while(len(ret)<10):
		if(dev.inWaiting()==0): blockcount+=1
		else: ret += dev.read(1)
		if(blockcount>1000):
			speech.speakMessage('hotplug error')
			return ''
		
	return ret

			

class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	"""papenmeier_serial braille display driver.
	"""
	name = "papenmeier_serial"
	description = _("Papenmeier BRAILLEX OLD")

	@classmethod
	def check(cls):
		"""should return false if there is a missing dependency"""
		return True

	
	def initTable(self):
		"""do not use braille builtin table"""
		table = []
		for i in range(0, self._numcells): table +=[1]
		self._dev.write(brl_out(512+self._offsetHorizontal, table))
	
	def __init__(self):
		"""Initializes braille display driver"""
		super(BrailleDisplayDriver, self).__init__()
		self._numcells=0
		self._r1next='r1a'
		self._lastkey=''
		self._baud=0
		self._dev=None
		self._port=None
		self._repeatcount=0

		
		#get path ti config file
		path = os.path.join(config.getConfigDirs()[0], "papenmeier_serial.ini")

		#read config file for serial port
		try:
			if(self._port == None):
				cfg = open(path, 'r')
				cfgline = cfg.readline().split("=")
				if(len(cfgline) == 2 and 'PORT' == cfgline[0]): self._port=cfgline[1]
				log.info(str(self._port)+' is connected to a braille display')
				cfg.close()
		except:
			log.info('config file not readable')
			
		try:
			if(self._port==None):
				cfg = open(path, 'w')
				self._port = 'COM1'
				cfg.write('PORT='+self._port)
				log.info(str(self._port)+' should be connected to a braille display')
				cfg.close()
		except:
			log.info('config file not writable')

		#try to connect to braille display        
		dic = -1
		for baud in [19200, 38400]:
			if(self._dev == None):
				self._dev = serial.Serial(self._port, baudrate=baud)
				self._dev.write(brl_auto_id())
				time.sleep(0.5)
				displaytype = brl_poll(self._dev)
				dic = -1
				if(len(displaytype)==10 and ord(displaytype[0])==STX and displaytype[1]=='I'):
					dic = ord(displaytype[2])
					if(baud == 19200): self._eab = False
					if(baud == 38400): self._eab = True
				print 'dic='+str(dic)
				if(dic == -1):
					self._dev.close()
					self._dev=None

					
		#set parameters for display                
		if(dic==2):
			self._numcells = 40
			self._offsetHorizontal = 0
			self._keymap = ['l1', 'l2', 'left', 'up', 'r2', 'dn', 'right', 'r1', 'reportf']
		elif(dic==1):
			self._offsetHorizontal = 0
			self._numcells = 40
		elif(dic==3):
			self._numcells = 80
			self._offsetHorizontal = 22
			self._keymap = ['l1', 'l2', 'r2', 'reportf', 'up', 'left', 'r1', 'right', 'dn', 'left2', 'up2', 'dn2', 'right2']
		elif(dic==3):
			self._numcells = 80
			self._offsetHorizontal = 0
		elif(dic==66):#//BRAILLEX EL 80
			self._offsetHorizontal = 2
			self._numcells = 80
		elif(dic==67):
			self._offsetHorizontal = 20
			self._numcells = 80
		elif(dic==65):
			self._offsetHorizontal = 0
			self._numcells = 66
		elif(dic==68):
			self._offsetHorizontal = 0
			self._numcells = 40
		elif(dic==69):
			self._offsetHorizontal = 0
			self._numcells = 32
		elif(dic==70):
			self._offsetHorizontal = 0
			self._numcells = 20
		else:
			raise Exception("No or unknown braille display found")

		
		#initialize display
		self.initTable()

		
		#start keyCheckTimer
		self._decodedkeys = []
		
		self._keyCheckTimer = wx.PyTimer(self._handleKeyPresses)
		self._keyCheckTimer.Start(KEY_CHECK_INTERVAL)
	      

	def terminate(self):
		"""free resources"""
		try:
			if(self._dev!=None):
				self._dev.close()
				self._dev = None
				self._keyCheckTimer.Stop()
				self._keyCheckTimer = None
			super(BrailleDisplayDriver, self).terminate()
		except:
			pass

	def _get_numCells(self):
		"""returns number of cells"""
		return self._numcells

	def display(self, cells):
		"""write data to braille display"""
		if(self._dev!=None):
			try:
				self._dev.write(brl_out(self._offsetHorizontal, cells))
			except:
				self._dev = None
				speech.speakMessage('braille display disconnected')
		

	def executeGesture(self,gesture):
		"""execute a gesture"""
		if(gesture.id=='r1'):  gesture.id = self.r1next()
		if(gesture.id=='r1,left'): gesture.id = 'left2'
		if(gesture.id=='r1,right'):  gesture.id = 'right2'
		if(gesture.id=='r1,up'):  gesture.id = 'up2'
		if(gesture.id=='r1,dn'):  gesture.id = 'dn2'
		if(gesture.id=='r2,left'): self.initTable()
		#here you can add other gesture types
		try:
			if(len(gesture.id)): inputCore.manager.executeGesture(gesture)
		except:
			print 'gesture'
	
		
	def _handleKeyPresses(self): #called by the keycheck timer
		"""if a button was pressed an input gesture is executed"""
		if(self._dev!=None):
			data = brl_poll2(self._dev)
			if(len(data)==10 and data[1]=='K'):
				pos = ord(data[2])*256+ord(data[3])
				pos = (pos-768)/3
				pressed = ord(data[6])
				keys = ord(data[8])
				self._repeatcount=0
				self.executeGesture(InputGesture(pos,pressed,keys,self))
			elif(len(data)==0):
				if(self._repeatcount==50):
					if(len(self._lastkey)): self.executeGesture(InputGesture(None, None, None, self))
					self._repeatcount=0
				else:
					self._repeatcount+=1
					
	
	def r1next(self):
		"""select next button for the braille display"""
		ret = self._r1next
		if(ret == 'r1a'): self._r1next = 'r1b'
		if(ret == 'r1b'): self._r1next = 'r1a'
		return ret

	#global gestures
	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(papenmeier_serial):left",),
			"braille_scrollForward": ("br(papenmeier_serial):right",),
			"braille_previousLine": ("br(papenmeier_serial):up",),
			"braille_nextLine": ("br(papenmeier_serial):dn",),
			"braille_routeTo": ("br(papenmeier_serial):route",),

			"navigatorObject_moveToFlatReviewAtObjectPosition": ("br(papenmeier_serial):r1a",),
			"braille_toggleTether": ("br(papenmeier_serial):r2",),
			"review_currentCharacter": ("br(papenmeier_serial):l1",),
			"navigatorObject_toFocus":("br(papenmeier_serial):r1b",),
			"navigatorObject_doDefaultAction": ("br(papenmeier_serial):l2",),
			"reportFormatting": ("br(papenmeier_serial):reportf",),

			"navigatorObject_previous": ("br(papenmeier_serial):left2",),
			"navigatorObject_next": ("br(papenmeier_serial):right2",),
			"navigatorObject_parent": ("br(papenmeier_serial):up2",),
			"navigatorObject_firstChild": ("br(papenmeier_serial):dn2",),

			"title": ("br(papenmeier_serial):l1,up",),
			"reportStatusLine": ("br(papenmeier_serial):l2,dn",),
			
		}
	})

#keys on display with eab
# 1 4 .......... 16 64
# 2 8 .......... 32 128


def brl_keyname2(keys):
	"""returns keyname for key index on displays with eab"""
	if(keys & 4 == 4): return 'l1'
	if(keys & 8 == 8): return 'l2'
	if(keys & 16 == 16): return 'r1'
	if(keys & 32 == 32): return 'r2'
	return ''



def brl_keyname(keyindex, driver):
	"""returns keyname for key index"""
	if(driver._eab):
		if(keyindex==-255): return "left"
		if(keyindex==-254): return "left2"
		if(keyindex==-253): return "up"
		if(keyindex==-252): return "up2"
		if(keyindex==-251): return "right"
		if(keyindex==-250): return "right2"
		if(keyindex==-249): return "dn"
		if(keyindex==-248): return "dn2"
		return ''
	else:
		#display does not have an eab, so the display specific table is used
		keyindex = keyindex+255
		if(keyindex>=0 and keyindex<len(driver._keymap)):
			return driver._keymap[keyindex]
		else: return ''

	

	

class InputGesture(braille.BrailleDisplayGesture):

	"""input gesture class for papenmeier_serial displays used only by the driver"""

	source = BrailleDisplayDriver.name
	
	def __init__(self, keyindex ,pressed ,keys, driver):
		super(InputGesture, self).__init__()
		self.id = ''
		if(keyindex==None):
			self.id = driver._lastkey
			return
		if(pressed==1 and keyindex>=0):
			self.routingIndex = keyindex-driver._offsetHorizontal
			self.id = "route"
			if(keyindex>255):
				braille.handler.routeTo(self.routingIndex-256)
				time.sleep(0.5)
				self.id = 'reportf'
				
			
		elif(pressed==0):
			k = brl_keyname(keyindex, driver)
			if(driver._lastkey!=k):
				if(driver._lastkey!=''): self.id=driver._lastkey+','+k
			else:
				self.id=k
				if(len(driver._decodedkeys) and len(k)): self.id = driver._decodedkeys[0]+","+k
				elif(len(driver._decodedkeys)==1): self.id = driver._decodedkeys[0]
				elif(len(driver._decodedkeys)==2): self.id = driver._decodedkeys[0]+','+driver._decodedkeys[1]
				driver._decodedkeys = [] 
			driver._lastkey = ''
		else:
			if(driver._lastkey == ''): driver._lastkey=brl_keyname(keyindex, driver)

		keys2=brl_keyname2(keys)
		if(len(keys2) and driver._eab): driver._decodedkeys += [keys2]

		

	def _get_id(self):
		"""return id of the display"""
		return 'papenmeier_serial' + str(self.id)

			
			